"""
Something like beancount's own API with some niceties.
"""

from __future__ import annotations

from decimal import Decimal
from functools import total_ordering
from typing import TYPE_CHECKING, Any, Literal

from attrs import evolve, field, frozen
from rpds import Queue

from alubia.exceptions import InvalidTransaction

if TYPE_CHECKING:
    from collections.abc import Iterable, Sequence
    from datetime import date
    from numbers import Number

    from attrs import Attribute


def _to_beancount_account_str(parts: Iterable[str]):
    return ":".join(parts)


_DEFAULT_WIDTH = 100


@frozen
class Transaction:
    """
    A beancount transaction.
    """

    date: date
    postings: Sequence[Posting] = field()
    payee: str

    @postings.validator  # type: ignore[reportAttributeAccessIssue]
    def _check(self, attribute: Attribute, value: Sequence[Posting]):  # type: ignore[reportUnknownParameterType]
        implicit = {posting for posting in value if posting.is_implicit}
        if len(implicit) > 1:
            raise InvalidTransaction(
                f"Multiple postings have implicit amounts: {implicit}",
            )

    def explicit(self):
        """
        A version of this transaction where all postings have explicit amounts.
        """
        if not self.postings:
            return self

        # FIXME: Amount to balance can have multiple currencies
        implicit: int | None = None
        total: Amount | Literal[0] = 0
        postings: list[Posting | None] = []
        for i, posting in enumerate(self.postings):
            if posting.is_implicit:
                assert implicit is None, "somehow multiple implicit postings??"
                implicit = i
                postings.append(None)
            else:
                total += posting.amount  # type: ignore[reportOperatorIssue]
                postings.append(posting)
        if implicit is None:
            return self

        postings[implicit] = -evolve(self.postings[implicit], amount=total)
        return evolve(self, postings=postings)

    def serialize(self, width: int = _DEFAULT_WIDTH):
        """
        Export this transaction to beancount's format.
        """
        lines = [f'{self.date} * "{self.payee}"']
        lines.extend(
            f"  {posting.serialize(width)}" for posting in self.postings
        )
        return "\n".join(lines)


@frozen
class Posting:
    """
    A leg of a transaction (i.e. a single account and amount).
    """

    _account: Account = field(alias="account", repr=str)
    amount: Amount | None = field(default=None)

    def __neg__(self):
        if self.amount is None:
            return NotImplemented
        return evolve(self, amount=-self.amount)

    @property
    def is_implicit(self):
        """
        Does this posting have an amount?
        """
        return self.amount is None

    def serialize(self, width: int):
        """
        Export this posting to beancount's line format.
        """
        amount = str(self.amount or "")
        padding = width - len(amount)
        return f"{self._account:<{padding}}{amount}"

    def transact(
        self,
        *postings: Posting | Account,
        **kwargs: Any,
    ) -> Transaction:
        """
        Create a transaction with this posting in it.
        """
        combined = [self]
        combined.extend(each.posting() for each in postings)  # type: ignore[reportArgumentType]
        return Transaction(postings=combined, **kwargs)

    def posting(self):
        """
        We are already one.
        """
        return self


@frozen
class Account:
    """
    A beancount account.
    """

    _parts: Queue[str] = field(
        alias="parts",
        converter=Queue,
        repr=_to_beancount_account_str,
    )
    _prefix: str = field(alias="prefix", default="")

    def __getattr__(self, name: str):
        """
        Get a child of this account if the name part is valid.
        """
        if not name[0].isupper():
            raise AttributeError(
                f"'{type(self).__name__}' object has no attribute '{name}'",
            )
        return self.child(name)

    def __getitem__(self, name: str):
        """
        Get a child of this account.
        """
        return self.child(name)

    def __invert__(self):
        """
        Mark this account flagged when it is part of a posting.
        """
        prefix = "" if self._prefix else "! "
        return evolve(self, prefix=prefix)

    def __format__(self, spec: str):
        return format(str(self), spec)

    def __str__(self):
        return f"{self._prefix}{_to_beancount_account_str(self._parts)}"

    def balance(self, date: date, amount: Amount, width: int = _DEFAULT_WIDTH):
        """
        Emit a balance assertion on this account.
        """
        without_prefix = width - 29
        return f"{date} balance {self:<{without_prefix}} {amount}"

    def child(self, name: str):
        """
        A child of this account.
        """
        return evolve(self, parts=self._parts.enqueue(name))

    def posting(self, **kwargs: Any) -> Posting:
        """
        A posting for this account.
        """
        return Posting(account=self, **kwargs)

    def transact[**P](self, *args: Posting, **kwargs: P.kwargs):  # type: ignore[reportGeneralTypeIssues]
        """
        Transact on this account as an implicit posting.
        """
        return self.posting().transact(*args, **kwargs)


@frozen
@total_ordering
class Amount:
    """
    A number of a specific commodity.
    """

    number: Decimal
    commodity: str

    @classmethod
    def from_str(cls, value: str) -> Amount:
        """
        Extract an amount from a string.
        """
        if value.startswith("(") and value.endswith(")"):
            return -cls.from_str(value[1:-1])

        rest = value[1:].replace(",", "")
        match value[0]:
            case "$":
                return cls(number=Decimal(rest), commodity="USD")
            case _:
                raise NotImplementedError(value)

    def __add__(self, other: Amount):
        if other.commodity != self.commodity:
            return NotImplemented
        return evolve(self, number=self.number + other.number)

    def __lt__(self, other: Amount):
        if other == 0:  # type: ignore[reportUnnecessaryComparison] um. wut?
            other = self.zero()
        elif (
            self.__class__ is not other.__class__
            or self.commodity != other.commodity
        ):
            return NotImplemented
        return self.number < other.number

    def __radd__(self, other: Number):
        if other != 0:  # type: ignore[reportUnnecessaryComparison] um. wut?
            return NotImplemented
        return self

    def __rmul__(self, other: Number):
        return evolve(self, number=other * self.number)  # type: ignore[reportOperatorIssue]

    def __neg__(self):
        return evolve(self, number=-self.number)

    def __str__(self):
        if self.number == self.number.to_integral():
            number = self.number
        else:
            number = self.number.quantize(Decimal("0.01"))
        return f"{number} {self.commodity}"

    def zero(self):
        """
        Zero in this commodity.
        """
        return evolve(self, number=Decimal(0))


Assets = Account(["Assets"])
Expenses = Account(["Expenses"])
Income = Account(["Income"])
Liabilities = Account(["Liabilities"])



"""
Ingestion helpers for reading in files.
"""

from __future__ import annotations

from csv import DictReader
from datetime import date
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from collections.abc import Iterable
    from pathlib import Path


def from_csv(
    path: Path,
    date_from: str = "Date",
    payee_from: str = "Description",
) -> Iterable[tuple[date, str, dict[str, Any]]]:
    """
    Partially parse a given csv path.
    """
    with path.open(newline="") as contents:
        reader = DictReader(contents)
        for row in reader:
            date_str = row.pop(date_from)
            row_date = date.strptime(date_str, "%m/%d/%Y")
            payee = row.pop(payee_from).strip()
            yield row_date, payee, row



from datetime import date
from decimal import Decimal

import pytest

from alubia.data import (
    Amount,
    Assets,
    Expenses,
    Income,
    Liabilities,
    Posting,
    Transaction,
)
from alubia.exceptions import InvalidTransaction

TODAY = date.today()
USD100 = Amount.from_str("$100.00")
USD200 = Amount.from_str("$200.00")


class TestAccount:
    def test_account_child(self):
        assert str(Expenses.Food.Meals) == "Expenses:Food:Meals"
        assert str(Assets.Bank.Checking) == "Assets:Bank:Checking"

    def test_account_invalid(self):
        with pytest.raises(AttributeError):
            Expenses.food

    def test_dynamic_account(self):
        account = Liabilities.Credit["visa".title()]
        assert str(account) == "Liabilities:Credit:Visa"

    def test_flagged_account(self):
        assert str(~Liabilities.Credit.Visa) == "! Liabilities:Credit:Visa"

    def test_posting(self):
        account = Assets.Bank.Checking
        posting = account.posting(amount=USD100)
        assert posting == Posting(account=account, amount=USD100)

    def test_transact(self):
        assert BANK.transact(
            Liabilities.Credit.Visa.posting(amount=USD100),
            date=date.today(),
            payee="Foo",
        ) == Transaction(
            payee="Foo",
            date=date.today(),
            postings=[
                BANK.posting(),
                Liabilities.Credit.Visa.posting(amount=USD100),
            ],
        )

    def test_flagged(self):
        assert str(~Assets.Bank.Checking) == "! Assets:Bank:Checking"


BANK = Assets.Bank.Checking


class TestPosting:
    def test_posting_negation(self):
        account = Assets.Bank.Checking
        posting = Posting(account=account, amount=USD100)
        assert -posting == Posting(account=account, amount=-USD100)

    def test_transact(self):
        posting = BANK.posting(amount=USD100)
        transaction = posting.transact(
            Liabilities.Credit.Visa.posting(),
            date=date.today(),
            payee="Foo Bar",
        )
        assert transaction == Transaction(
            payee="Foo Bar",
            date=date.today(),
            postings=[
                BANK.posting(amount=USD100),
                Liabilities.Credit.Visa.posting(),
            ],
        )

    def test_transact_bare_account(self):
        posting = BANK.posting(amount=USD100)
        transaction = posting.transact(
            Liabilities.Credit.Visa,
            payee="Baz Quux",
            date=date.today(),
        )
        assert transaction == Transaction(
            date=date.today(),
            payee="Baz Quux",
            postings=[
                BANK.posting(amount=USD100),
                Liabilities.Credit.Visa.posting(),
            ],
        )

    def test_default_amount(self):
        assert Posting(account=BANK, amount=None) == Posting(account=BANK)

    def test_implicit(self):
        implicit = Posting(account=Assets.Bank.Checking)
        assert implicit.is_implicit

    def test_explicit(self):
        explicit = Posting(account=Assets.Bank.Checking, amount=USD100)
        assert not explicit.is_implicit


class TestTransaction:
    def test_explicit_two_postings_one_implicit(self):
        tx = Transaction(
            date=TODAY,
            payee="",
            postings=[
                Posting(account=Assets.Cash, amount=USD100),
                Posting(account=Income.Salary),
            ],
        )
        assert tx.explicit() == Transaction(
            date=TODAY,
            payee="",
            postings=[
                Posting(account=Assets.Cash, amount=USD100),
                Posting(account=Income.Salary, amount=-USD100),
            ],
        )

    def test_explicit_multiple_postings(self):
        tx = Transaction(
            date=TODAY,
            payee="",
            postings=[
                Posting(account=Assets.Cash, amount=USD100),
                Posting(account=Assets.Cash, amount=USD200),
                Posting(account=Income.Salary),
            ],
        )
        assert tx.explicit() == Transaction(
            date=TODAY,
            payee="",
            postings=[
                Posting(account=Assets.Cash, amount=USD100),
                Posting(account=Assets.Cash, amount=USD200),
                Posting(account=Income.Salary, amount=-(USD100 + USD200)),
            ],
        )

    def test_explicit_all_postings_have_amounts(self):
        tx = Transaction(
            date=TODAY,
            payee="",
            postings=[
                Posting(account=Assets.Cash, amount=USD100),
                Posting(account=Income.Salary, amount=-USD100),
            ],
        )
        assert tx.explicit() is tx

    def test_missing_amounts(self):
        p1 = Posting(account=Assets.Cash)
        p2 = Posting(account=Income.Salary)
        with pytest.raises(InvalidTransaction):
            Transaction(payee="", date=TODAY, postings=[p1, p2])


class TestAmount:
    def test_from_str_dollars(self):
        amount = Amount.from_str("$100.00")
        assert amount == Amount(commodity="USD", number=Decimal(100))

    def test_from_str_accounting_notation(self):
        amount = Amount.from_str("($100.00)")
        assert amount == Amount(commodity="USD", number=Decimal(-100))

    def test_from_str_invalid(self):
        with pytest.raises(NotImplementedError):
            Amount.from_str("asdf")

    def test_zero(self):
        zero_usd = Amount.from_str("$100.00").zero()
        assert zero_usd == Amount(commodity="USD", number=Decimal(0))

    def test_add(self):
        assert USD100 + USD200 == Amount(commodity="USD", number=Decimal(300))

    def test_neg(self):
        assert -USD100 == Amount(commodity="USD", number=Decimal(-100))

    def test_gt(self):
        assert USD100 > 0

    def test_lt(self):
        assert USD100 < USD200

    def test_str_exact_dollar(self):
        amount = Amount(commodity="USD", number=Decimal(100))
        assert str(amount) == "100 USD"

    def test_str_cents(self):
        amount = Amount(commodity="USD", number=Decimal("100.23"))
        assert str(amount) == "100.23 USD"

    def test_str_quantized_round_down(self):
        amount = Amount(commodity="USD", number=Decimal("100.23456"))
        assert str(amount) == "100.23 USD"

    def test_str_quantized_round_up(self):
        amount = Amount(commodity="USD", number=Decimal("100.23566"))
        assert str(amount) == "100.24 USD"
